@@ -1,5 +1,5 @@
 # ----------------------------------------------------------------------------
-# Copyright (c) 2016-2021, QIIME 2 development team.
+# Copyright (c) 2016-2022, QIIME 2 development team.
 #
 # Distributed under the terms of the Modified BSD License.
 #
@@ -16,6 +16,7 @@ from q2_types.feature_table import FeatureTable, Frequency
 
 import q2_dada2
 from q2_dada2 import DADA2Stats, DADA2StatsFormat, DADA2StatsDirFmt
+import q2_dada2._examples as ex
 
 _POOL_OPT = {'pseudo', 'independent'}
 _CHIM_OPT = {'pooled', 'consensus', 'none'}
@@ -46,6 +47,7 @@ plugin.methods.register_function(
                 'chimera_method': qiime2.plugin.Str %
                 qiime2.plugin.Choices(_CHIM_OPT),
                 'min_fold_parent_over_abundance': qiime2.plugin.Float,
+                'allow_one_off': qiime2.plugin.Bool,
                 'n_threads': qiime2.plugin.Int,
                 'n_reads_learn': qiime2.plugin.Int,
                 'hashed_feature_ids': qiime2.plugin.Bool},
@@ -98,6 +100,11 @@ plugin.methods.register_function(
             'than or equal to 1 (i.e. parents should be more abundant than '
             'the sequence being tested). This parameter has no effect if '
             'chimera_method is "none".'),
+        'allow_one_off': (
+            'Bimeras that are one-off from exact are also '
+            'identified if the `allow_one_off` argument is True.'
+            'If True, a sequence will be identified as bimera if it is one '
+            'mismatch or indel away from an exact bimera.'),
         'n_threads': ('The number of threads to use for multithreaded '
                       'processing. If 0 is provided, all available cores will '
                       'be used.'),
@@ -122,7 +129,10 @@ plugin.methods.register_function(
     },
     name='Denoise and dereplicate single-end sequences',
     description=('This method denoises single-end sequences, dereplicates '
-                 'them, and filters chimeras.')
+                 'them, and filters chimeras.'),
+    examples={
+        'denoise_single': ex.denoise_single
+    }
 )
 
 
@@ -143,6 +153,7 @@ plugin.methods.register_function(
                 'chimera_method': qiime2.plugin.Str %
                 qiime2.plugin.Choices(_CHIM_OPT),
                 'min_fold_parent_over_abundance': qiime2.plugin.Float,
+                'allow_one_off': qiime2.plugin.Bool,
                 'n_threads': qiime2.plugin.Int,
                 'n_reads_learn': qiime2.plugin.Int,
                 'hashed_feature_ids': qiime2.plugin.Bool},
@@ -216,6 +227,11 @@ plugin.methods.register_function(
             'than or equal to 1 (i.e. parents should be more abundant than '
             'the sequence being tested). This parameter has no effect if '
             'chimera_method is "none".'),
+        'allow_one_off': (
+            'Bimeras that are one-off from exact are also '
+            'identified if the `allow_one_off` argument is True'
+            'If True, a sequence will be identified as bimera if it is one '
+            'mismatch or indel away from an exact bimera.'),
         'n_threads': ('The number of threads to use for multithreaded '
                       'processing. If 0 is provided, all available cores will '
                       'be used.'),
@@ -242,7 +258,10 @@ plugin.methods.register_function(
     },
     name='Denoise and dereplicate paired-end sequences',
     description=('This method denoises paired-end sequences, dereplicates '
-                 'them, and filters chimeras.')
+                 'them, and filters chimeras.'),
+    examples={
+        'denoise_paired': ex.denoise_paired
+    }
 )
 
 
@@ -259,6 +278,7 @@ plugin.methods.register_function(
                 'chimera_method': qiime2.plugin.Str %
                 qiime2.plugin.Choices(_CHIM_OPT),
                 'min_fold_parent_over_abundance': qiime2.plugin.Float,
+                'allow_one_off': qiime2.plugin.Bool,
                 'n_threads': qiime2.plugin.Int,
                 'n_reads_learn': qiime2.plugin.Int,
                 'hashed_feature_ids': qiime2.plugin.Bool},
@@ -291,7 +311,7 @@ plugin.methods.register_function(
                    'longer than this value. If 0 is provided no reads will '
                    'be removed based on length.',
         'pooling_method': 'The method used to pool samples for denoising. '
-                          '"independent": Samples are denoised indpendently. '
+                          '"independent": Samples are denoised independently. '
                           '"pseudo": The pseudo-pooling method is used to '
                           'approximate pooling of samples. In short, samples '
                           'are denoised independently once, ASVs detected '
@@ -313,6 +333,11 @@ plugin.methods.register_function(
             'than or equal to 1 (i.e. parents should be more abundant than '
             'the sequence being tested). This parameter has no effect if '
             'chimera_method is "none".',
+        'allow_one_off': (
+            'Bimeras that are one-off from exact are also '
+            'identified if the `allow_one_off` argument is True. '
+            'If True, a sequence will be identified as bimera if it is one '
+            'mismatch or indel away from an exact bimera.'),
         'n_threads': 'The number of threads to use for multithreaded '
                      'processing. If 0 is provided, all available cores will '
                      'be used.',
@@ -341,6 +366,140 @@ plugin.methods.register_function(
 )
 
 
+plugin.methods.register_function(
+    function=q2_dada2.denoise_ccs,
+    inputs={'demultiplexed_seqs': SampleData[SequencesWithQuality]},
+    parameters={'front': qiime2.plugin.Str,
+                'adapter': qiime2.plugin.Str,
+                'max_mismatch': qiime2.plugin.Int,
+                'indels': qiime2.plugin.Bool,
+                'trunc_len': qiime2.plugin.Int,
+                'trim_left': qiime2.plugin.Int,
+                'max_ee': qiime2.plugin.Float,
+                'trunc_q': qiime2.plugin.Int,
+                'min_len': qiime2.plugin.Int,
+                'max_len': qiime2.plugin.Int,
+                'pooling_method': qiime2.plugin.Str %
+                qiime2.plugin.Choices(_POOL_OPT),
+                'chimera_method': qiime2.plugin.Str %
+                qiime2.plugin.Choices(_CHIM_OPT),
+                'min_fold_parent_over_abundance': qiime2.plugin.Float,
+                'allow_one_off': qiime2.plugin.Bool,
+                'n_threads': qiime2.plugin.Int,
+                'n_reads_learn': qiime2.plugin.Int,
+                'hashed_feature_ids': qiime2.plugin.Bool},
+    outputs=[('table', FeatureTable[Frequency]),
+             ('representative_sequences', FeatureData[Sequence]),
+             ('denoising_stats', SampleData[DADA2Stats])],
+    input_descriptions={
+        'demultiplexed_seqs': 'The single-end demultiplexed PacBio CCS '
+                              'sequences to be denoised.'
+    },
+    parameter_descriptions={
+        'front': 'Sequence of an adapter ligated to the 5\' end. '
+                 'The adapter and any preceding bases are trimmed. '
+                 'Can contain IUPAC ambiguous nucleotide codes. '
+                 'Note, primer direction is 5\' to 3\'. '
+                 'Primers are removed before trim and filter step. '
+                 'Reads that do not contain the primer are discarded. '
+                 'Each read is re-oriented if the reverse complement of '
+                 'the read is a better match to the provided primer sequence. '
+                 'This is recommended for PacBio CCS reads, which come in a '
+                 'random mix of forward and reverse-complement orientations.',
+        'adapter': 'Sequence of an adapter ligated to the 3\' end. '
+                   'The adapter and any preceding bases are trimmed. '
+                   'Can contain IUPAC ambiguous nucleotide codes. '
+                   'Note, primer direction is 5\' to 3\'. '
+                   'Primers are removed before trim and filter step. '
+                   'Reads that do not contain the primer are discarded.',
+        'max_mismatch': 'The number of mismatches to tolerate when matching '
+                        'reads to primer sequences '
+                        '- see http://benjjneb.github.io/dada2/ '
+                        'for complete details.',
+        'indels': 'Allow insertions or deletions of bases when '
+                  'matching adapters. Note that primer matching can '
+                  'be significantly slower, currently about 4x slower',
+        'trunc_len': 'Position at which sequences should be truncated due to '
+                     'decrease in quality. This truncates the 3\' end of the '
+                     'of the input sequences, which will be the bases that '
+                     'were sequenced in the last cycles. Reads that are '
+                     'shorter than this value will be discarded. If 0 is '
+                     'provided, no truncation or length filtering will be '
+                     'performed. Note: Since Pacbio CCS sequences were '
+                     'normally with very high quality scores, '
+                     'there is no need to truncate the Pacbio CCS sequences.',
+        'trim_left': 'Position at which sequences should be trimmed due to '
+                     'low quality. This trims the 5\' end of the '
+                     'of the input sequences, which will be the bases that '
+                     'were sequenced in the first cycles.',
+        'max_ee': 'Reads with number of expected errors higher than this '
+                  'value will be discarded.',
+        'trunc_q': 'Reads are truncated at the first instance of a quality '
+                   'score less than or equal to this value. If the resulting '
+                   'read is then shorter than `trunc_len`, it is discarded.',
+        'min_len': 'Remove reads with length less than minLen. '
+                   'minLen is enforced after trimming and truncation.'
+                   ' For 16S Pacbio CCS, suggest 1000.',
+        'max_len': 'Remove reads prior to trimming or truncation which are '
+                   'longer than this value. If 0 is provided no reads will '
+                   'be removed based on length. '
+                   'For 16S Pacbio CCS, suggest 1600.',
+        'pooling_method': 'The method used to pool samples for denoising. '
+                          '"independent": Samples are denoised indpendently. '
+                          '"pseudo": The pseudo-pooling method is used to '
+                          'approximate pooling of samples. In short, samples '
+                          'are denoised independently once, ASVs detected '
+                          'in at least 2 samples are recorded, and samples '
+                          'are denoised independently a second time, but '
+                          'this time with prior knowledge of the recorded '
+                          'ASVs and thus higher sensitivity to those ASVs.',
+        'chimera_method': 'The method used to remove chimeras. '
+                          '"none": No chimera removal is performed. '
+                          '"pooled": All reads are pooled prior to chimera '
+                          'detection. "consensus": Chimeras are detected in '
+                          'samples individually, and sequences found chimeric '
+                          'in a sufficient fraction of samples are removed.',
+        'min_fold_parent_over_abundance':
+            'The minimum abundance of potential parents of a sequence being '
+            'tested as chimeric, expressed as a fold-change versus the '
+            'abundance of the sequence being tested. Values should be greater '
+            'than or equal to 1 (i.e. parents should be more abundant than '
+            'the sequence being tested). Suggest 3.5. '
+            'This parameter has no effect if chimera_method is "none".',
+        'allow_one_off': (
+            'Bimeras that are one-off from exact are also '
+            'identified if the `allow_one_off` argument is True. '
+            'If True, a sequence will be identified as bimera if it is one '
+            'mismatch or indel away from an exact bimera.'),
+        'n_threads': 'The number of threads to use for multithreaded '
+                     'processing. If 0 is provided, all available cores will '
+                     'be used.',
+        'n_reads_learn': 'The number of reads to use when training the '
+                         'error model. Smaller numbers will result in a '
+                         'shorter run time but a less reliable error model.',
+        'hashed_feature_ids': 'If true, the feature ids in the resulting '
+                              'table will be presented as hashes of the '
+                              'sequences defining each feature. The hash '
+                              'will always be the same for the same sequence '
+                              'so this allows feature tables to be merged '
+                              'across runs of this method. You should only '
+                              'merge tables if the exact same parameters are '
+                              'used for each run.'
+    },
+    output_descriptions={
+        'table': 'The resulting feature table.',
+        'representative_sequences': 'The resulting feature sequences. Each '
+                                    'feature in the feature table will be '
+                                    'represented by exactly one sequence.'
+    },
+    name='Denoise and dereplicate single-end Pacbio CCS',
+    description='This method denoises single-end Pacbio CCS sequences, '
+                'dereplicates them, and filters chimeras. '
+                'Tutorial and workflow: '
+                'https://github.com/benjjneb/LRASManuscript'
+)
+
+
 plugin.register_formats(DADA2StatsFormat, DADA2StatsDirFmt)
 plugin.register_semantic_types(DADA2Stats)
 plugin.register_semantic_type_to_format(
