@@ -1,5 +1,5 @@
 # ----------------------------------------------------------------------------
-# Copyright (c) 2016-2021, QIIME 2 development team.
+# Copyright (c) 2016-2022, QIIME 2 development team.
 #
 # Distributed under the terms of the Modified BSD License.
 #
@@ -16,7 +16,7 @@ from q2_types.per_sample_sequences import (
     SingleLanePerSampleSingleEndFastqDirFmt,
     SingleLanePerSamplePairedEndFastqDirFmt)
 
-from q2_dada2 import denoise_single, denoise_paired, denoise_pyro
+from q2_dada2 import denoise_single, denoise_paired, denoise_pyro, denoise_ccs
 from q2_dada2._denoise import _check_featureless_table
 
 
@@ -24,6 +24,13 @@ def _sort_seqs(seqs):
     return sorted(list(seqs), key=lambda x: x.metadata['id'])
 
 
+class TestExamples(TestPluginBase):
+    package = 'q2_dada2.tests'
+
+    def test_examples(self):
+        self.execute_examples()
+
+
 class TestDenoiseSingle(TestPluginBase):
     package = 'q2_dada2.tests'
 
@@ -73,6 +80,14 @@ class TestDenoiseSingle(TestPluginBase):
                          _sort_seqs(exp_rep_seqs))
         self.assertEqual(md, exp_md)
 
+    def test_mixed_barcodes_and_ids(self):
+        demux_seqs = SingleLanePerSamplePairedEndFastqDirFmt(
+            self.get_data_path('mixed_barcodes_and_ids'), 'r')
+
+        denoise_paired(demux_seqs, 150, 150)
+
+        self.assertTrue(True)
+
     def test_all_reads_filtered(self):
         with self.assertRaisesRegex(ValueError, 'filter'):
             denoise_single(self.demux_seqs, 10000)
@@ -333,5 +348,35 @@ class TestUtils(TestPluginBase):
             _check_featureless_table(fp)
 
 
+class TestDenoiseCCS(TestPluginBase):
+    package = 'q2_dada2.tests'
+
+    def setUp(self):
+        super().setUp()
+        self.demux_seqs = SingleLanePerSampleSingleEndFastqDirFmt(
+            self.get_data_path('sample_seqs_ccs'), 'r')
+
+    def test_defaults(self):
+        with open(self.get_data_path('expected/ccs-default.tsv')) as fh:
+            exp_table = biom.Table.from_tsv(fh, None, None, lambda x: x)
+        exp_rep_seqs = list(
+            skbio.io.read(self.get_data_path('expected/ccs-default.fasta'),
+                          'fasta', constructor=skbio.DNA))
+        for seq in exp_rep_seqs:
+            del seq.metadata['description']
+        exp_md = qiime2.Metadata.load(
+            self.get_data_path('expected/ccs-default-stats.tsv'))
+
+        table, rep_seqs, md = denoise_ccs(
+                                  self.demux_seqs,
+                                  front="AGRGTTYGATYMTGGCTCAG",
+                                  adapter="RGYTACCTTGTTACGACTT",)
+
+        self.assertEqual(table, exp_table)
+        self.assertEqual(_sort_seqs(rep_seqs),
+                         _sort_seqs(exp_rep_seqs))
+        self.assertEqual(md, exp_md)
+
+
 if __name__ == '__main__':
     unittest.main()
